#!/usr/bin/env python

"""
Translate the AST of a Python program into a more interpretable representation.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import micropython.inspect
from micropython.rsvp import *
from micropython.common import *
import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class TranslateError(NodeProcessingError): pass

class Label:

    "A reference to a location."

    def __init__(self, number, location=None):
        self.number = number
        self.location = location

    def __repr__(self):
        return "Label(%r, location=%r)" % (self.number, self.location)

# Program visitors.

class Translation(ASTVisitor):

    "A translated module."

    supported_optimisations = ["constant_storage", "known_target", "self_access"]

    def __init__(self, module, importer, optimisations=None):

        """
        Initialise the translation with an inspected 'module', the 'importer'
        and optional 'optimisations'. See the 'supported_optimisations'
        attribute of this class for permitted values.
        """

        ASTVisitor.__init__(self)
        self.visitor = self
        self.module = module

        # Global program dependencies.

        self.importer = importer
        self.objtable = self.importer.get_object_table()
        self.paramtable = self.importer.get_parameter_table()
        self.builtins = self.importer.modules.get("__builtins__")

        # Desired optimisations.

        self.optimisations = set(optimisations or [])

        # The current unit being translated.

        self.unit = None

        # Wiring within the code.

        self.labels = {}
        self.label_number = 0
        self.loop_labels = []
        self.exception_labels = []

        # The code itself. This is limited to the code for a particular block
        # being processed.

        self.code = None
        self.temp_position = 0

    def calculate_stack_usage(self):
        max_stack_usage = 0
        stack_usage = 0

        for op in self.code:
            stack_usage += op.stack_usage
            max_stack_usage = max(max_stack_usage, stack_usage)

        self.unit.stack_usage = max_stack_usage

    def get_module_code(self):

        "Return the top-level module code."

        self.unit = self.module
        self.code = []
        self.temp_position = self.unit.stack_local_usage

        if self.module.module is not None:
            self.dispatch(self.module.module)

        self.calculate_stack_usage()
        return self.code

    def get_code(self, unit):

        "Return the code for the given 'unit'."

        self.unit = unit
        self.code = []
        self.temp_position = self.unit.stack_local_usage

        if unit.node is not None:
            self.dispatch(unit.node)

        self.calculate_stack_usage()
        return self.code

    def __repr__(self):
        return "Translation(%r)" % self.module

    def get_scope(self, name):
        if self.unit.has_key(name):
            return "local"
        elif self.module.has_key(name):
            return "global"
        else:
            return "builtins"

    # Code writing methods.

    def new_label(self):

        "Return a new label object for use with set_label."

        number = self.label_number
        label = Label(number)
        self.labels[label] = label
        self.label_number += 1
        return label

    def set_label(self, label):

        """
        Set the location of 'label' to that within the entire image: the
        location within the code combined with location of the code unit.
        """

        label.location = len(self.code) + self.unit.code_location

    def get_loop_labels(self):
        return self.loop_labels[-1]

    def add_loop_labels(self, next_label, exit_label):
        self.loop_labels.append((next_label, exit_label))

    def drop_loop_labels(self):
        self.loop_labels.pop()

    def get_exception_labels(self):
        return self.exception_labels[-1]

    def add_exception_labels(self, handler_label, exit_label):
        self.exception_labels.append((handler_label, exit_label))

    def drop_exception_labels(self):
        self.exception_labels.pop()

    def reserve_temp(self, n):
        temp_position = self.temp_position
        self.temp_position += n
        self.unit.stack_temp_usage = max(self.unit.stack_temp_usage, self.temp_position)
        return temp_position

    def discard_temp(self, n):
        self.temp_position -= n

    def new_op(self, op):

        "Add 'op' to the generated code."

        self.code.append(op)

    def replace_op(self, op):

        "Replace the last added instruction with 'op'."

        self.code[-1] = op

    def remove_ops(self, n):

        "Remove the last 'n' instructions."

        del self.code[-n:]

    def last_ops(self, n):

        "Return the last 'n' added instructions in reverse chronological order."

        ops = self.code[-n:]
        ops.reverse()
        return ops

    def last_op(self):

        "Return the last added instruction."

        try:
            return self.code[-1]
        except IndexError:
            return None

    # Internal helper methods.

    def _visitAttr(self, node, classes):

        """
        Visit the attribute-related 'node', generating instructions based on the
        given 'classes'.
        """

        self.dispatch(node.expr)
        self._generateAttr(node, node.attrname, classes)

    def _generateAttr(self, node, attrname, classes):

        """
        Generate code for the access to 'attrname' using the given 'classes'.
        """

        AddressInstruction, AttrInstruction, AttrIndexInstruction = classes

        last = self.last_op()

        # Where the last operation (defining the attribute owner) yields a
        # constant...

        if self._have_constant_input(0):

            # Optimise away the constant storage if appropriate.

            if self._optimise_constant_storage(AddressInstruction, 1):
                return

            # Get the details of the access.

            target = last.attr.value
            target_name = target.full_name()

            try:
                table_entry = self.objtable.table[target_name]
            except KeyError:
                raise TranslateError(self.module.full_name(), node,
                    "No object entry exists for target %r." % target_name)

            try:
                pos = table_entry[attrname]
            except KeyError:
                raise TranslateError(self.module.full_name(), node,
                    "No attribute entry exists for name %r in target %r." % (attrname, target_name))

            if isinstance(target, micropython.inspect.Instance):
                self.replace_op(AttrInstruction(pos))
            else:
                self.replace_op(AddressInstruction(pos))

        # Where the last operation involves the special 'self' name, check to
        # see if the attribute is acceptably positioned and produce a direct
        # access to the attribute.

        elif self._optimise_self_access(attrname, AttrInstruction):
            pass

        # Otherwise, perform a normal operation.

        else:
            try:
                index = self.objtable.get_index(attrname)
            except self.objtable.TableError:
                raise TranslateError(self.module.full_name(), node,
                    "No attribute entry exists for name %r." % attrname)

            self.new_op(AttrIndexInstruction(index))

    def _startCallFunc(self):

        "Record the location of the invocation."

        self.new_op(MakeFrame()) # records the start of the frame

    def _generateCallFunc(self, args, node):

        # NOTE: Only simple cases are used for optimisations.

        target, context = self._optimise_known_target()

        # Where a target is known and has a known context, avoid generating any
        # first argument. Instance methods do not have a known target since they
        # are accessed via an instance whose identity cannot generally be known
        # at compile-time.

        if context is None:
            continue_label = self.new_label()
            self.new_op(LoadContext())
            self.new_op(CheckContext())
            self.new_op(JumpIfTrue(continue_label))
            self.dispatch(compiler.ast.Name("TypeError"))
            self.new_op(RaiseException())
            self.set_label(continue_label)
        else:
            pass # NOTE: Class methods should be supported.

        # Evaluate the arguments.

        employed_positions = set()
        extra_keywords = []

        for frame_pos, arg in enumerate(args):

            # Handle positional and keyword arguments separately.

            if isinstance(arg, compiler.ast.Keyword):

                # Optimise where the target is known now.

                if target is not None:

                    # Find the parameter table entry for the target.

                    target_name = target.full_name()

                    # Look for a callable with the precise target name.

                    table_entry = self.paramtable.table[target_name]

                    # Look the name up in the parameter table entry.

                    try:
                        pos = table_entry[arg.name]

                    # Where no position is found, this could be an extra keyword
                    # argument.

                    except KeyError:
                        extra_keywords.append(arg)
                        continue

                    # Test for illegal conditions.

                    if pos in employed_positions:
                        raise TranslateError(self.module.full_name(), node,
                            "Keyword argument %r overwrites parameter %r." % (arg.name, pos))

                    employed_positions.add(pos)

                    # Add space for arguments appearing before this one.

                    if frame_pos < pos:
                        self.new_op(ReserveFrame(pos - frame_pos))

                    # Generate code for the keyword and the positioning
                    # operation.

                    self.dispatch(arg.expr)

                    # If the position corresponds to the current frame element,
                    # skip generating the store instruction.

                    if frame_pos > pos:
                        self.new_op(StoreFrame(pos))

                # Otherwise, generate the code needed to obtain the details of
                # the parameter location.

                else:

                    # Combine the target details with the name to get the location.
                    # See the access method on the List class.

                    try:
                        paramindex = self.paramtable.get_index(arg.name)

                    # Where no position is found, this could be an extra keyword
                    # argument.

                    except self.paramtable.TableError:
                        extra_keywords.append(arg)
                        continue

                    # Generate code for the keyword and the positioning
                    # operation.

                    self.dispatch(arg.expr)
                    self.new_op(StoreFrameIndex(paramindex))

                    # use (callable+0)+paramindex+table
                    # checks embedded offset against (callable+0)
                    # moves the top of stack to frame+position

            else:
                self.dispatch(arg)
                employed_positions.add(frame_pos)

        frame_pos = len(args)

        # NOTE: If any extra keywords were identified, generate them now.
        # NOTE: This needs to populate a ** argument dictionary.
        #
        # for arg in extra_keywords:
        #     const = self.module.constant_values[arg.name]
        #     self.new_op(LoadConst(const))
        #     self.dispatch(arg.expr)
        #     frame_pos += 1

        # NOTE: Somehow, the above needs to be combined with * and ** arguments.

        # Either test for a complete set of arguments.

        if target is not None:

            # Make sure that enough arguments have been given.

            nargs_max = len(target.positional_names)
            ndefaults = len(target.defaults)
            nargs_min = nargs_max - ndefaults

            for i in range(0, nargs_min):
                if i not in employed_positions:
                    raise TranslateError(self.module.full_name(), node,
                        "Argument %r not supplied for %r: need at least %d arguments." % (i+1, target.name, nargs_min))

            nargs = len(args)

            if nargs > nargs_max and not target.has_star and not target.has_dstar:
                raise TranslateError(self.module.full_name(), node,
                    "Too many arguments for %r: need at most %d arguments." % (target.name, nargs_max))

            # Where defaults are involved, put them into the frame.
            # Here, we use negative index values to visit the right hand end of
            # the defaults list.

            for pos in range(nargs_min, nargs_max):
                if pos not in employed_positions:
                    self.new_op(LoadConst(target))
                    self.new_op(LoadAttr(target.default_attrs[pos - nargs_min]))

                # If the position corresponds to the current frame element,
                # skip generating the instruction.

                if frame_pos != pos:
                    self.new_op(StoreFrame(pos))

                frame_pos += 1

        # Or generate instructions to do this at run-time.
        # NOTE: CheckFrame has to check the number of arguments and to fill in
        # NOTE: defaults.

        else:
            self.new_op(CheckFrame())

    def _endCallFunc(self):

        "Make the invocation and tidy up afterwards."

        self.new_op(LoadCallable()) # uses the start of the frame to get the callable
        self.new_op(JumpWithFrame())

        # NOTE: Exception handling required.

        self.new_op(DropFrame())

    def _visitName(self, node, classes):

        """
        Visit the name-related 'node', generating instructions based on the
        given 'classes'.
        """

        name = node.name
        scope = self.get_scope(name)
        #print self.module.name, node.lineno, name, scope
        self._generateName(name, scope, classes, node)

    def _generateName(self, name, scope, classes, node):

        """
        Generate code for the access to 'name' in 'scope' using the given
        'classes', and using the given 'node' as the source of the access.
        """

        NameInstruction, AddressInstruction = classes

        if self._optimise_constant_storage(NameInstruction, 0):
            return

        if scope == "local":
            unit = self.unit
            if isinstance(unit, micropython.inspect.Function):
                self.new_op(NameInstruction(unit.all_locals()[name]))
            elif isinstance(unit, micropython.inspect.Class):
                self.new_op(AddressInstruction(unit.all_class_attributes()[name]))
            elif isinstance(unit, micropython.inspect.Module):
                self.new_op(AddressInstruction(unit.module_attributes()[name]))
            else:
                raise TranslateError(self.module.full_name(), node, "Program unit %r has no local %r." % (unit, name))

        elif scope == "global":
            globals = self.module.module_attributes()
            if globals.has_key(name):
                self.new_op(AddressInstruction(globals[name]))
            else:
                raise TranslateError(self.module.full_name(), node, "Module %r has no attribute %r." % (self.module, name))

        else:
            self.new_op(AddressInstruction(self._get_builtin(name, node)))

    def _get_builtin(self, name, node):
        if self.builtins is not None:
            try:
                return self.builtins[name]
            except KeyError:
                raise TranslateError(self.module.full_name(), node, "No __builtins__ definition is available for name %r." % name)
        else:
            raise TranslateError(self.module.full_name(), node, "No __builtins__ module is available for name %r." % name)

    # Optimisation tests.

    def _should_optimise_constant_storage(self):
        return "constant_storage" in self.optimisations

    def _should_optimise_known_target(self):
        return "known_target" in self.optimisations

    def _should_optimise_self_access(self):
        return "self_access" in self.optimisations

    def _have_constant_input(self, n):
        last = self.last_ops(n+1)
        return len(last) > n and (isinstance(last[n], LoadAddress) and last[n].attr.assignments == 1 or
            isinstance(last[n], LoadConst))

    def _have_known_target(self):
        return self._have_constant_input(0)

    def _have_self_input(self):
        last = self.last_op()
        return isinstance(self.unit, micropython.inspect.Function) and \
            self.unit.is_method() and isinstance(last, LoadName) and \
            last.attr.name == "self"

    # Optimisation methods. See the supported_optimisations class attribute.

    def _optimise_constant_storage(self, instruction, n):

        """
        Where this operation should store a constant into a target which is
        also constant, optimise away both operations.
        """

        if self._should_optimise_constant_storage() and \
            instruction in (StoreAddress, StoreName) and \
            self._have_constant_input(n) and \
            (n == 0 or self._have_constant_input(n-1)):

            self.remove_ops(n+1)
            return 1
        else:
            return 0

    def _optimise_known_target(self):

        """
        Where the target of an invocation is known, provide information about it
        and its context. If a class is being invoked and the conditions are
        appropriate, get information about the specific initialiser.
        """

        if self._should_optimise_known_target() and self._have_known_target():
            last = self.last_op()
            target = last.attr.value
            context = last.attr.parent

            # Handle calls to classes.

            if isinstance(target, micropython.inspect.Class):
                target = target.get_instantiator()
                context = micropython.inspect.Instance()

                # A special context is chosen to avoid generating unnecessary
                # context loading and checking instructions.

        else:
            target = None
            context = None

        return target, context

    def _optimise_self_access(self, attrname, instruction):

        """
        Where the provided 'attrname' accesses an attribute which occupies the
        same position in all possible objects which can be accessed, generate an
        'instruction' accessing the attribute directly.
        """

        if self._should_optimise_self_access() and self._have_self_input() and \
            not self.unit.is_relocated(attrname):

            attr = self.unit.parent.all_attributes()[attrname]
            self.new_op(instruction(attr))
            return 1
        else:
            return 0

    # Visitor methods.

    def default(self, node, *args):
        raise TranslateError(self.module.full_name(), node, "Node class %r is not supported." % node.__class__)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def _visitBinary(self, node, left_method, right_method):

        """
        _t1 = node.left
        _t2 = node.right
        try:
            _t1.__add__(_t2)
        except AttributeError:
            _t2.__radd__(_t1)
        """

        right_label = self.new_label()
        end_label = self.new_label()

        # NOTE: Decide on temporary storage access.

        temp_position = self.reserve_temp(2)

        self.dispatch(node.left)
        self.new_op(StoreTemp(temp_position))
        self.dispatch(node.right)
        self.new_op(StoreTemp(temp_position + 1))

        # Left method.

        self._startCallFunc()
        self.new_op(LoadTemp(temp_position))
        self._generateAttr(node, left_method, (LoadAddress, LoadAttr, LoadAttrIndex))
        self.new_op(LoadTemp(temp_position)) # Explicit context as first argument.
        self.new_op(LoadTemp(temp_position + 1))
        self._endCallFunc()

        self.dispatch(compiler.ast.Name("AttributeError"))
        self.new_op(CheckException())
        self.new_op(JumpIfFalse(end_label))

        # Right method.

        self.set_label(right_label)
        self._startCallFunc()
        self.new_op(LoadTemp(temp_position + 1))
        self._generateAttr(node, right_method, (LoadAddress, LoadAttr, LoadAttrIndex))
        self.new_op(LoadTemp(temp_position + 1)) # Explicit context as first argument.
        self.new_op(LoadTemp(temp_position))
        self._endCallFunc()

        self.set_label(end_label)

        self.discard_temp(2)

    def visitAdd(self, node):
        self._visitBinary(node, "__add__", "__radd__")

    def visitAnd(self, node): pass

    def visitAssert(self, node): pass

    def visitAssign(self, node):
        self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)

    def visitAssAttr(self, node):
        self._visitAttr(node, (StoreAddress, StoreAttr, StoreAttrIndex))

    def visitAssList(self, node): pass

    def visitAssName(self, node):
        self._visitName(node, (StoreName, StoreAddress))

    visitAssTuple = visitAssList

    def visitAugAssign(self, node): pass

    def visitBackquote(self, node): pass

    def visitBitand(self, node): pass

    def visitBitor(self, node): pass

    def visitBitxor(self, node): pass

    def visitBreak(self, node):
        next_label, exit_label = self.get_loop_labels()
        self.new_op(Jump(exit_label))

    def visitCallFunc(self, node):

        """
        Evaluate positional arguments, evaluate and store keyword arguments in
        the correct location, then invoke the function.
        """

        # Mark the frame, evaluate the target, generate the call.

        self._startCallFunc()
        self.dispatch(node.node)
        self._generateCallFunc(node.args, node)
        self._endCallFunc()

    def visitClass(self, node):
        unit = self.unit
        self.unit = node.unit
        self.unit.code_location = self.module.code_location # class body code is not independently addressable
        self.dispatch(node.code)
        self.unit = unit

    def visitCompare(self, node):

        """
        self.dispatch(node.expr)
        for op_name, next_node in compare.ops:
            methods = self.comparison_methods[op_name]
            if methods is not None:
                # Generate method call using evaluated argument and next node.
            else:
                # Deal with the special operators.
            # Provide short-circuiting.
        """

    def visitConst(self, node):
        const = self.module.constant_values[node.value]
        self.new_op(LoadConst(const))

    def visitContinue(self, node):
        next_label, exit_label = self.get_loop_labels()
        self.new_op(Jump(next_label))

    def visitDecorators(self, node): pass

    def visitDict(self, node): pass

    def visitDiscard(self, node):
        self.dispatch(node.expr)

    def visitDiv(self, node):
        self._visitBinary(node, "__div__", "__rdiv__")

    def visitEllipsis(self, node): pass

    def visitExec(self, node): pass

    def visitExpression(self, node): pass

    def visitFloorDiv(self, node):
        self._visitBinary(node, "__floordiv__", "__rfloordiv__")

    def visitFor(self, node):
        exit_label = self.new_label()
        next_label = self.new_label()
        else_label = self.new_label()

        # Get the "list" to be iterated over, obtain its iterator.

        self._startCallFunc()
        self.dispatch(node.list)
        self._generateAttr(node, "__iter__", (LoadAddress, LoadAttr, LoadAttrIndex))
        self._generateCallFunc([], node)
        self._endCallFunc()

        # Iterator on stack.

        # In the loop...

        self.set_label(next_label)

        # Use the iterator to get the next value.

        self._startCallFunc()
        self.new_op(Duplicate())
        self._generateAttr(node, "next", (LoadAddress, LoadAttr, LoadAttrIndex))
        self._generateCallFunc([], node)
        self._endCallFunc()

        # Test for StopIteration.

        self.dispatch(compiler.ast.Name("StopIteration"))
        self.new_op(CheckException())
        if node.else_ is not None:
            self.new_op(JumpIfTrue(else_label))
        else:
            self.new_op(JumpIfTrue(exit_label))

        # Assign to the target.

        self.dispatch(node.assign)

        # Process the body with the current next and exit points.

        self.add_loop_labels(next_label, exit_label)
        self.dispatch(node.body)
        self.drop_loop_labels()

        # Repeat the loop.

        self.new_op(Jump(next_label))

        # Produce the "else" section.

        if node.else_ is not None:
            self.set_label(exit_label)
            self.dispatch(node.else_)

        # Pop the iterator.

        self.set_label(exit_label)
        self.new_op(Pop())

    def visitFrom(self, node): pass

    def visitFunction(self, node):

        # Only store the name when visiting this node from outside.

        if self.unit is not node.unit:
            self.new_op(LoadConst(node.unit))
            self._visitName(node, (StoreName, StoreAddress))

            # Generate the default initialisation code.

            for attr, default in zip(node.unit.default_attrs, node.unit.defaults):
                self.dispatch(default)
                self.new_op(StoreAddress(attr))

        # Visiting of the code occurs when get_code is invoked on this node.

        else:
            self.dispatch(node.code)
            if not isinstance(self.last_op(), Return):
                self.dispatch(compiler.ast.Name("None"))
                self.new_op(Return())

    def visitGenExpr(self, node): pass

    def visitGenExprFor(self, node): pass

    def visitGenExprIf(self, node): pass

    def visitGenExprInner(self, node): pass

    def visitGetattr(self, node):
        self._visitAttr(node, (LoadAddress, LoadAttr, LoadAttrIndex))

    def visitGlobal(self, node): pass

    def visitIf(self, node):
        first = 1
        exit_label = self.new_label()

        for test, body in node.tests + [(None, node.else_)]:
            if body is None:
                break
            if not first:
                self.set_label(next_label)
            if test is not None:
                self.dispatch(test)
                next_label = self.new_label()
                self.new_op(JumpIfFalse(next_label))
            self.dispatch(body)
            self.new_op(Jump(exit_label))
            first = 0

        self.set_label(exit_label)

    def visitImport(self, node): pass

    def visitInvert(self, node): pass

    def visitKeyword(self, node): pass

    def visitLambda(self, node): pass

    def visitLeftShift(self, node): pass

    def visitList(self, node): pass

    def visitListComp(self, node): pass

    def visitListCompFor(self, node): pass

    def visitListCompIf(self, node): pass

    def visitMod(self, node):
        self._visitBinary(node, "__mod__", "__rmod__")

    def visitModule(self, node):
        self.dispatch(node.node)

    def visitMul(self, node):
        self._visitBinary(node, "__mul__", "__rmul__")

    def visitName(self, node):
        self._visitName(node, (LoadName, LoadAddress))

    def visitNot(self, node): pass

    def visitOr(self, node): pass

    def visitPass(self, node): pass

    def visitPower(self, node): pass

    def visitPrint(self, node): pass

    def visitPrintnl(self, node): pass

    def visitRaise(self, node): pass

    def visitReturn(self, node):
        if node.value is not None:
            self.dispatch(node.value)
        else:
            self.dispatch(compiler.ast.Name("None"))
        self.new_op(Return())

    def visitRightShift(self, node): pass

    def visitSlice(self, node): pass

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)

    def visitSub(self, node):
        self._visitBinary(node, "__sub__", "__rsub__")

    def visitSubscript(self, node): pass

    def visitTryExcept(self, node):

        """
        Enter try block.
        Dispatch to code.
        
        """

        exit_label = self.new_label()
        handler_label = self.new_label()

        self.add_exception_labels(handler_label, exit_label)

        # Try...
        # Produce the code, then jump to the exit.

        self.dispatch(node.body)
        self.new_op(Jump(exit_label))

        # Start of handlers.

        self.set_label(handler_label)
        for name, assignment, handler in node.handlers:
            next_label = self.new_label()

            # Test the given exception against the current exception.

            if name is not None:
                self.dispatch(name)
                self.new_op(CheckException())
                self.new_op(JumpIfFalse(next_label))

            # Handle assignment to exception variable.

            if assignment is not None:
                self.dispatch(assignment)

            # Produce the handler code, then jump to the exit.

            self.dispatch(handler)
            self.new_op(Jump(exit_label))

            self.set_label(next_label)

        # Unhandled exceptions.

        self.new_op(RaiseException())

        # After exception 

        self.set_label(exit_label)

        # Optional else clause.

        if node.else_ is not None:
            self.dispatch(node.else_)

        self.drop_exception_labels()

    def visitTryFinally(self, node): pass

    def visitTuple(self, node): pass

    def visitUnaryAdd(self, node): pass

    def visitUnarySub(self, node): pass

    def visitWhile(self, node):
        exit_label = self.new_label()
        next_label = self.new_label()
        else_label = self.new_label()

        self.set_label(next_label)
        self.dispatch(node.test)
        if node.else_ is not None:
            self.new_op(JumpIfFalse(else_label))
        else:
            self.new_op(JumpIfFalse(exit_label))

        self.add_loop_labels(next_label, exit_label)

        self.dispatch(node.body)
        self.new_op(Jump(next_label))

        if node.else_ is not None:
            self.set_label(else_label)
            self.dispatch(node.else_)

        self.set_label(exit_label)
        self.drop_loop_labels()

    def visitWith(self, node): pass

    def visitYield(self, node): pass

    # Useful data.

    comparison_methods = {
        "==" : ("__eq__", "__ne__"),
        "!=" : ("__ne__", "__eq__"),
        "<" : ("__lt__", "__gt__"),
        "<=" : ("__le__", "__ge__"),
        ">=" : ("__ge__", "__le__"),
        ">" : ("__gt__", "__lt__"),
        "is" : None,
        "is not" : None,
        "in" : None,
        "not in" : None
        }

# vim: tabstop=4 expandtab shiftwidth=4
